@@ -35,6 +35,7 @@ SUGGESTION_UPDATE_INTERVAL = 200
 
 xux = lambda x: f'{x} {x.upper()}'
 PROJECT_FILETYPES = [('Project Files', xux('.json')), ('All Files', '.*')]
+PYTHON_FILETYPES = [('Python Files', xux('.py')), ('All Files', '.*')]
 IMAGE_FILETYPES = [('Images', xux('.png .jpg .jpeg')), ('All Files', '.*')]
 
 color_enabled = False
@@ -99,6 +100,11 @@ def _process_print_queue():
                 pass # throwing would break print queue
     root.after(33, _process_print_queue)
 
+def basename_noext(path: str) -> str:
+    file = os.path.basename(path)
+    p = file.find('.')
+    return (file[:p] if p >= 0 else file).strip()
+
 def get_white_nonwhite(line: str) -> Tuple[str, str]:
     i = 0
     while i < len(line) and line[i].isspace():
@@ -323,7 +329,7 @@ class Content(tk.Frame):
         self.project.grid(row = 0, column = 0, sticky = tk.NSEW)
         self.display.grid(row = 0, column = 1, sticky = tk.NSEW)
 
-        self.grid_columnconfigure(0, weight = 4, uniform = 'content')
+        self.grid_columnconfigure(0, weight = 5, uniform = 'content')
         self.grid_columnconfigure(1, weight = 3, uniform = 'content')
         self.grid_rowconfigure(0, weight = 1)
 
@@ -419,10 +425,52 @@ class BlocksList(tk.Frame):
 
 class Imports:
     RAW_INFO = [
-        ['time', 'time'],
-        ['math', 'math'],
-        ['random', 'random'],
-        ['numpy', 'np'],
+        ['time', 'time', '''
+            The `time` package brings in several utilities related to measuring time.
+
+            Here is a list of especially useful functions:
+             - `time.sleep(t)` lets your scripts wait a certain number of seconds before resuming.
+             - `time.time()` gives you the current timestamp. You can subtract two times to get the time interval in seconds.
+        '''],
+        ['math', 'math', '''
+            The `math` module brings in many common mathematical functions such as `sin`, `cos`, `tan`,
+            inverse trig, exponentiation, and many more specialized tools.
+
+            This is useful if you need to compute a handful of operations in your scripts.
+            However, if you need to do a lot of number crunching, consider using the `numpy` package instead.
+        '''],
+        ['random', 'random', '''
+            The `random` module allows you to generate random numbers.
+
+            Here is a list of especially useful functions:
+             - `random.random()` gives you a float (decimal) number between 0.0 and 1.0 (not including 1.0).
+             - `random.randrange(a,b)` gives you a random integer in `range(a,b)`.
+            Just like the `range` function, you can pass a single argument to default the first to 0.
+        '''],
+        ['numpy', 'np', '''
+            The `numpy` package is a highly specialized tool for doing a lot of number crunching very very quickly.
+            By convention, `numpy` is typically imported as `np` for brevity.
+
+            Most `numpy` functions work on arrays (lists). To make a `numpy` array, simply use `np.array(some_list)`.
+            `numpy` arrays perform their operations in batches.
+            For instance, if you have two `numpy` arrays `p` and `q`, each of which is an array of three numbers like 3D coordinates,
+            you can simply write `p+q` and numpy will add the components together and give you a new array.
+            Many operations are supported, for instance `np.sin(p + 2 * q**2)` will square all the elements of `q`, double them,
+            add them to the elements of `p`, then take the `sin` of each element and give the result as a new array.
+
+            `numpy` arrays are actually multi-dimensional arrays.
+            A 1D array is a "normal" array (list).
+            A 2D array is a matrix.
+            In higher dimensions (and in general) these are called tensors, and are commonly used in advanced applications like machine learning.
+            If you pass `np.array()` a list of lists, you will get an N-dimensional array (tensor) with the same structure.
+
+            Python supports list slicing. For instance `p[3:5]` gets the sublist of elements with index 3 up to (but not including) 5.
+            However, python lists do not support higher-dimensional slicing (even if they are lists of lists).
+            But `numpy` arrays allow this, with each dimension separated by a comma.
+            For instance if `p` is a 5x5 matrix, `p[1,2:4]` gives row index 1 elements 2 up to 4, and `p[1:-1,1:-1]` gives you the 3x3 matrix in the middle.
+
+            As you can see, `numpy` lets you do a lot of work very quickly and without much typing.
+        '''],
     ]
     def __init__(self, *, on_update = None):
         self.packages = {}
@@ -430,7 +478,8 @@ class Imports:
             self.packages[item[0]] = {
                 'tkvar': tk.BooleanVar(),
                 'ident': item[1],
-                'code': f'import {item[0]}' if item[0] == item[1] else f'import {item[0]} as {item[1]}'
+                'info': clean_docstring(normalize_strip(item[2])),
+                'code': f'import {item[0]}' if item[0] == item[1] else f'import {item[0]} as {item[1]}',
             }
         
         self.images = {}
@@ -468,6 +517,11 @@ class Imports:
 class ProjectEditor(tk.Frame):
     def __init__(self, parent):
         super().__init__(parent)
+
+        # this is randomly generated and is valid while the ide is open accross multiple program runs.
+        # it is not stored in project files or exports because conflicting ids would break messaging.
+        self.project_id = common.generate_proj_id()
+
         self.turtle_index = 0
         self.editors: List[CodeEditor] = []
         self.show_blocks_tkvar = tk.BooleanVar()
@@ -558,49 +612,49 @@ class ProjectEditor(tk.Frame):
             self.notebook.add(editor, text = name)
             self.editors.append(editor)
     
-    def get_full_script(self) -> str:
+    def get_full_script(self, *, is_export: bool = False) -> str:
         scripts = []
         for editor in self.editors:
-            scripts.append(editor.get_script())
+            scripts.append(editor.get_script(is_export = is_export))
             scripts.append('\n\n')
         scripts.append('start_project()')
         return ''.join(scripts)
 
     DEFAULT_GLOBAL_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef function_name():\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef function_name():\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
-        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
 
-        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message_type\')\ndef function_name(): # add arguments to receive values\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message_type\', fields = 4, go = 2, here = 7)' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in list_goes_here:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
     ]
     DEFAULT_STAGE_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef function_name(self):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef function_name(self):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef function_name(self, x, y):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
-        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
 
-        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message_type\')\ndef function_name(self): # add arguments to receive values\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message_type\', fields = 4, go = 2, here = 7)' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in list_goes_here:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(self, msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
     ]
     DEFAULT_TURTLE_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef function_name(self):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/onstartclone.png', 'scale': 1, 'replace': '@onstartclone\ndef function_name(self):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef function_name(self):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef function_name(self, x, y):\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onstartclone.png', 'scale': 1, 'replace': '@onstartclone\ndef my_onstartclone(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
-        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
         
-        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message_type\')\ndef function_name(self): # add arguments to receive values\n    pass' },
-        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message_type\', fields = 4, go = 2, here = 7)' },
-        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in list_goes_here:\n    pass' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(self, msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
+        { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
     ]
     DEFAULT_PROJECT = {
         'global_blocks': DEFAULT_GLOBAL_BLOCKS,
@@ -1022,27 +1076,30 @@ class CodeEditor(ScrolledText):
         edit_line, edit_col = self.total_pos()
         docs = script.help(edit_line, edit_col)
 
-        def get_docstring(item):
-            mapped = FULL_NAME_DOC_REMAPS.get(item.full_name)
-            if mapped is not None:
-                return mapped
-
-            if item.column is not None and item.column > 0 and item.get_line_code()[item.column - 1] == '.':
-                mapped = PROP_DOC_REMAPS.get(item.name)
+        def get_docstring(items) -> str:
+            res = []
+            for item in items:
+                mapped = FULL_NAME_DOC_REMAPS.get(item.full_name)
                 if mapped is not None:
                     return mapped
 
-            desc = item.description
-            if desc.startswith('keyword') or desc.startswith('instance'):
-                return ''
+                if item.column is not None and item.column > 0 and item.get_line_code()[item.column - 1] == '.':
+                    mapped = PROP_DOC_REMAPS.get(item.name)
+                    if mapped is not None:
+                        return mapped
 
-            return item.docstring()
-        docs = [get_docstring(x) for x in docs]
-        docs = [x for x in docs if x] # don't show empty items
+                desc = item.description
+                if desc.startswith('keyword') or desc.startswith('instance'):
+                    continue
+
+                x = item.docstring()
+                if x:
+                    res.append(x)
+            return '\n\n----------\n\n'.join(res)
+        docs = get_docstring(docs)
 
         if docs: # if nothing to show, don't change the display
-            raw = '\n\n----------\n\n'.join(docs)
-            clean = clean_docstring(raw)
+            clean = clean_docstring(docs)
             content.display.docs.set_text(clean)
 
     def show_suggestion(self, script = None):
@@ -1165,18 +1222,18 @@ class CodeEditor(ScrolledText):
 class GlobalEditor(CodeEditor):
     BASE_PREFIX = '''
 import netsblox
-nb = netsblox.Client()
-'A connection to NetsBlox, which allows you to use services and RPCs from python.'
-netsblox.turtle._set_window_size(1280, 720)
 from netsblox.turtle import *
 from netsblox.concurrency import *
-import netsblox.common as _common
+nb = netsblox.Client(proj_name = """$proj_name""", proj_id = $proj_id)
+'A connection to NetsBlox, which allows you to use services and RPCs from python.'
+netsblox.turtle._set_window_size(1280, 720)
+netsblox.turtle._turtle.title(f'NetsBlox-Python - {nb.get_public_id()}')
+setup_stdio()
 setup_yielding()
 import time as _time
 def _yield_(x):
     _time.sleep(0)
     return x
-setup_stdio()
 
 '''.lstrip()
     BASE_PREFIX_LINES = 14
@@ -1190,8 +1247,11 @@ setup_stdio()
         super().__init__(parent, blocks = GlobalEditor.blocks)
         self.set_text(value)
 
-    def get_script(self):
-        return GlobalEditor.prefix + self.text.get('1.0', 'end-1c')
+    def get_script(self, *, is_export: bool = False):
+        pre = GlobalEditor.prefix
+        pre = pre.replace('$proj_name', main_menu.project_name)
+        pre = pre.replace('$proj_id', 'None' if is_export else f'\'{content.project.project_id}\'')
+        return pre + self.text.get('1.0', 'end-1c')
 
 class StageEditor(CodeEditor):
     prefix_lines = 3
@@ -1202,7 +1262,7 @@ class StageEditor(CodeEditor):
         self.name = name
         self.set_text(value)
 
-    def get_script(self):
+    def get_script(self, *, is_export: bool = False):
         raw = self.text.get('1.0', 'end-1c')
         return f'@netsblox.turtle.stage\nclass {self.name}(netsblox.turtle.StageBase):\n    pass\n{indent(raw)}\n{self.name} = {self.name}()'
 
@@ -1215,7 +1275,7 @@ class TurtleEditor(CodeEditor):
         self.name = name
         self.set_text(value)
 
-    def get_script(self):
+    def get_script(self, *, is_export: bool = False):
         raw = self.text.get('1.0', 'end-1c')
         return f'@netsblox.turtle.turtle\nclass {self.name}(netsblox.turtle.TurtleBase):\n    pass\n{indent(raw)}\n{self.name} = {self.name}()'
 
@@ -1338,11 +1398,24 @@ class MainMenu(tk.Menu):
 
         submenu = tk.Menu(self, **MENU_STYLE)
         imp = content.project.imports
-        for pkg, item in imp.packages.items():
+        imp_packages = list(imp.packages.items())
+        for pkg, item in imp_packages:
             label = pkg if pkg == item['ident'] else f'{pkg} ({item["ident"]})'
             submenu.add_checkbutton(label = label, variable = item['tkvar'], command = imp.batch_update)
         self.add_cascade(label = 'Imports', menu = submenu)
 
+        def get_info_shower(submenu):
+            def show_info(e):
+                entries = len(imp.packages)
+                entry = int(e.y / (submenu.winfo_reqheight() / entries))
+                if entry < 0 or entry >= entries: return
+
+                pkg, item = imp_packages[entry]
+                info = item['info']
+                if info: content.display.docs.set_text(item['info'])
+            return show_info
+        submenu.bind('<Motion>', get_info_shower(submenu))
+
         self.images_dropdown = tk.Menu(self, **MENU_STYLE)
         self.add_cascade(label = 'Images', menu = self.images_dropdown)
 
@@ -1364,7 +1437,11 @@ class MainMenu(tk.Menu):
     @project_path.setter
     def project_path(self, p):
         self._project_path = p
-        root.title(f'NetsBlox-Python - {"Untitled" if p is None else p}')
+        self._project_name = 'untitled' if p is None else basename_noext(p)
+        root.title(f'NetsBlox-Python - {"untitled" if p is None else p}')
+    @property
+    def project_name(self):
+        return self._project_name
 
     def save(self, save_dict = None) -> bool:
         if self.project_path is not None:
@@ -1388,10 +1465,10 @@ class MainMenu(tk.Menu):
         return False
 
     def export_as(self) -> None:
-        p = filedialog.asksaveasfilename(defaultextension = '.py')
+        p = filedialog.asksaveasfilename(filetypes = PYTHON_FILETYPES, defaultextension = '.py')
         if type(p) is str and p: # despite the type hints, above returns empty tuple on cancel
             try:
-                res = transform.add_yields(content.project.get_full_script())
+                res = transform.add_yields(content.project.get_full_script(is_export = True))
                 with open(p, 'w') as f:
                     f.write(res)
             except Exception as e:
